/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
/* eslint-disable @typescript-eslint/no-non-null-assertion */
import { strict as assert } from "assert";
import * as fs from "fs";
import { walkAllChildSegments } from "../mergeTreeNodeWalk.js";
import { toRemovalInfo } from "../mergeTreeNodes.js";
import { MergeTreeDeltaType, ReferenceType } from "../ops.js";
import { TextSegment } from "../textSegment.js";
import { _dirname } from "./dirname.cjs";
import { TestClientLogger } from "./testClientLogger.js";
export const removeRange = (client, opStart, opEnd) => client.removeRangeLocal(opStart, opEnd);
export const obliterateRange = (client, opStart, opEnd) => client.obliterateRangeLocal(opStart, opEnd);
export const annotateRange = (client, opStart, opEnd) => client.annotateRangeLocal(opStart, opEnd, { client: client.longClientId });
export const insertAtRefPos = (client, opStart, opEnd, random) => {
    const segs = [];
    // gather all the segments at the pos, including removed segments
    walkAllChildSegments(client.mergeTree.root, (seg) => {
        const pos = client.getPosition(seg);
        if (pos >= opStart) {
            if (pos <= opStart) {
                segs.push(seg);
                return true;
            }
            return false;
        }
        return true;
    });
    if (segs.length > 0) {
        const text = client.longClientId.repeat(random.integer(1, 3));
        const seg = random.pick(segs);
        const lref = client.createLocalReferencePosition(seg, toRemovalInfo(seg) ? 0 : random.integer(0, seg.cachedLength - 1), toRemovalInfo(seg)
            ? ReferenceType.SlideOnRemove
            : random.pick([
                ReferenceType.Simple,
                ReferenceType.SlideOnRemove,
                ReferenceType.Transient,
            ]), undefined);
        return client.insertAtReferencePositionLocal(lref, TextSegment.make(text));
    }
};
export const insert = (client, _start, _end, random) => {
    // Note: the _start param is generated using exclusive range. This provides more coverage by allowing
    // insertion at the end.
    const start = random.integer(0, client.getLength());
    const text = client.longClientId.repeat(random.integer(1, 3));
    return client.insertTextLocal(start, text);
};
export function doOverRange(range, defaultGrowthFunc, doAction) {
    let lastCurrent = Number.NaN;
    for (let current = range.min; current <= range.max; current = (range.growthFunc ?? defaultGrowthFunc)(current)) {
        // let growth funcs be simple
        // especially around 0 and 1
        // if the value didn't change,
        // just increment it
        if (current === lastCurrent) {
            current++;
        }
        if (current <= range.max) {
            lastCurrent = current;
            doAction(current);
        }
    }
}
export function resolveRange(range, defaultGrowthFunc) {
    const results = [];
    doOverRange(range, range.growthFunc ?? defaultGrowthFunc, (num) => {
        results.push(num);
    });
    return results;
}
export function resolveRanges(ranges, defaultGrowthFunc) {
    return Object.entries(ranges)
        .filter(([_, value]) => isConfigRange(value))
        .map(([key, value]) => [key, resolveRange(value, defaultGrowthFunc)])
        .reduce((prev, [key, resolvedRange]) => {
        prev[key] = resolvedRange;
        return prev;
    }, {});
}
function isConfigRange(t) {
    return typeof t === "object" && typeof t.min === "number" && typeof t.max === "number";
}
export function doOverRanges(ranges, doAction) {
    const rangeEntries = Object.entries(ranges).filter(([_, value]) => isConfigRange(value));
    const doOverRangesHelper = (selections) => {
        if (selections.length === rangeEntries.length) {
            const selectionsObj = {};
            for (const [key, value] of selections) {
                selectionsObj[key] = value;
            }
            const description = selections.map(([key, value]) => `${key}:${value}`).join("_");
            doAction(selectionsObj, description);
        }
        else {
            const [key, value] = rangeEntries[selections.length];
            doOverRange(value, value.growthFunc ?? ranges.growthFunc, (selection) => {
                selections.push([key, selection]);
                doOverRangesHelper(selections);
                selections.pop();
            });
        }
    };
    doOverRangesHelper([]);
}
export const replayResultsPath = `${_dirname}/../../src/test/results`;
export function runMergeTreeOperationRunner(random, startingSeq, clients, minLength, config, apply = applyMessages) {
    let seq = startingSeq;
    const results = [];
    doOverRange(config.opsPerRoundRange, config.growthFunc, (opsPerRound) => {
        if (config.incrementalLog) {
            console.log(`MinLength: ${minLength} Clients: ${clients.length} Ops: ${opsPerRound} Seq: ${seq}`);
        }
        for (let round = 0; round < config.rounds; round++) {
            const initialText = clients[0].getText();
            const logger = new TestClientLogger(clients, `Clients: ${clients.length} Ops: ${opsPerRound} Round: ${round}`);
            const messageData = generateOperationMessagesForClients(random, seq, clients, logger, opsPerRound, minLength, config.operations);
            const msgs = messageData.map((md) => md[0]);
            seq = apply(seq, messageData, clients, logger, random);
            const resultText = logger.validate();
            results.push({
                initialText,
                resultText,
                msgs,
                seq,
            });
            logger.dispose();
        }
    });
    if (config.resultsFilePostfix !== undefined) {
        const resultsFilePath = `${replayResultsPath}/len_${minLength}-clients_${clients.length}-${config.resultsFilePostfix}`;
        fs.writeFileSync(resultsFilePath, JSON.stringify(results, undefined, 4));
    }
    return seq;
}
export function generateOperationMessagesForClients(random, startingSeq, clients, logger, opsPerRound, minLength, operations) {
    const minimumSequenceNumber = startingSeq;
    let tempSeq = startingSeq * -1;
    const messages = [];
    for (let i = 0; i < opsPerRound; i++) {
        // pick a client greater than 0, client 0 only applies remote ops
        // and is our baseline
        const client = clients[random.integer(1, clients.length - 1)];
        const len = client.getLength();
        const sg = client.peekPendingSegmentGroups();
        let op;
        if (len === 0 || len < minLength) {
            const text = client.longClientId.repeat(random.integer(1, 3));
            op = client.insertTextLocal(random.integer(0, len), text);
        }
        else {
            let opIndex = random.integer(0, operations.length - 1);
            const start = random.integer(0, len - 1);
            const end = random.integer(start + 1, len);
            for (let y = 0; y < operations.length && op === undefined; y++) {
                op = operations[opIndex](client, start, end, random);
                opIndex++;
                opIndex %= operations.length;
            }
        }
        if (op !== undefined) {
            // Pre-check to avoid logger.toString() in the string template
            if (sg === client.peekPendingSegmentGroups()) {
                assert.notEqual(sg, client.peekPendingSegmentGroups(), `op created but segment group not enqueued.${logger}`);
            }
            const message = client.makeOpMessage(op, --tempSeq);
            message.minimumSequenceNumber = minimumSequenceNumber;
            messages.push([
                message,
                client.peekPendingSegmentGroups(op.type === MergeTreeDeltaType.GROUP ? op.ops.length : 1),
            ]);
        }
    }
    return messages;
}
export function generateClientNames() {
    const clientNames = [];
    function addClientNames(startChar, count) {
        const startCode = startChar.charCodeAt(0);
        for (let i = 0; i < count; i++) {
            clientNames.push(String.fromCharCode(startCode + i));
        }
    }
    addClientNames("A", 26);
    addClientNames("a", 26);
    addClientNames("0", 17);
    return clientNames;
}
export function applyMessages(startingSeq, messageData, clients, logger) {
    let seq = startingSeq;
    try {
        // log and apply all the ops created in the round
        // eslint-disable-next-line @typescript-eslint/prefer-for-of
        for (let i = 0; i < messageData.length; i++) {
            const [message] = messageData[i];
            message.sequenceNumber = ++seq;
            clients.forEach((c) => c.applyMsg(message));
        }
    }
    catch (e) {
        throw logger.addLogsToError(e);
    }
    return seq;
}
//# sourceMappingURL=mergeTreeOperationRunner.js.map